from pyemv import ac, kd, tools


def test_generate_ac_emv_arpc1() -> None:
    r"""
    Test generate AC with EMV padding (\x80 padding).

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using ICC Master Key
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key.
    pan = "9901234567890123"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "1DA5"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "0995"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text, ac.PaddingType.EMV)
    assert arqc.hex().upper() == "4B46013359B7A58B"

    # ARPC Method 1 using ICC Master Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(icc_mk, arqc, arpc_rc)
    assert arpc.hex().upper() == "F8C9CECAABD55AD1"


def test_generate_ac_default_emv_arpc1() -> None:
    r"""
    Test generate AC with default EMV padding (\x80 padding).

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using ICC Master Key
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key.
    pan = "9901234567890123"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "1DA5"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "0995"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text)
    assert arqc.hex().upper() == "4B46013359B7A58B"

    # ARPC Method 1 using ICC Master Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(icc_mk, arqc, arpc_rc)
    assert arpc.hex().upper() == "F8C9CECAABD55AD1"


def test_generate_ac_visa_aprc1() -> None:
    r"""
    Test generate AC with Visa padding (\x00 padding).

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using ICC Master Key
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "9901234567890123"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "1DA5"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "0995"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF01")
    arqc = ac.generate_ac(sk_ac, cipher_text, ac.PaddingType.VISA)
    assert arqc.hex().upper() == "2E141C6BC4A20DA8"

    # ARPC Method 1 using ICC Master Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(icc_mk, arqc, arpc_rc)
    assert arpc.hex().upper() == "16A49AAB314B9262"


def test_generate_ac_visa_aprc1_no_padding_required() -> None:
    r"""
    Test generate AC with Visa padding (\x00 padding).
    However, no padding is required. The data is already multiple of 8.

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using ICC Master Key
    ARPC Method = 1
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "9901234567890123"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "1DA5"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "0995"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text, ac.PaddingType.VISA)
    assert arqc.hex().upper() == "922F3E83125EB46B"

    # ARPC Method 1 using ICC Master Key
    arpc_rc = bytes.fromhex("0000")
    arpc = ac.generate_arpc_1(icc_mk, arqc, arpc_rc)
    assert arpc.hex().upper() == "8AE6E836084B0E80"


def test_generate_arpc2() -> None:
    r"""
    Test generate ARPC using method 2.

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key
    ARPC Method = 2
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "9901234567890123"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "1DA5"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "0995"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text, ac.PaddingType.EMV)
    assert arqc.hex().upper() == "4B46013359B7A58B"

    # ARPC Method 2 using ICC Master Key
    csu = bytes.fromhex("00000000")
    prop_auth_data = bytes.fromhex("1234567890ABCDEF")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu, prop_auth_data)
    assert arpc.hex().upper() == "4962B76C"


def test_generate_arpc2_no_prop_auth_data() -> None:
    r"""
    Test generate ARPC using method 2 without prop auth data.

    Master Key Derivation = Option A
    ARQC verification using Common Session Key Derivation
    ARPC generation using Common Session Key
    ARPC Method = 2
    """
    # Verify issuer master key check digits
    iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    assert tools.key_check_digits(iss_mk, 2).hex().upper() == "08D7"

    # Derive ICC master key
    pan = "9901234567890123"
    psn = "45"
    icc_mk = kd.derive_icc_mk_a(iss_mk, pan, psn)
    assert tools.key_check_digits(icc_mk, 2).hex().upper() == "1DA5"

    # Verify AC session key
    # Common Session Key Derivation Option
    r = bytes.fromhex("1234567890123456")
    sk_ac = kd.derive_common_sk(icc_mk, r)
    assert tools.key_check_digits(sk_ac, 2).hex().upper() == "0995"

    # ARQC validation using Session Key
    cipher_text = bytes.fromhex("0123456789ABCDEF0123456789ABCDEF")
    arqc = ac.generate_ac(sk_ac, cipher_text, ac.PaddingType.EMV)
    assert arqc.hex().upper() == "4B46013359B7A58B"

    # ARPC Method 2 using ICC Master Key
    csu = bytes.fromhex("00000000")
    arpc = ac.generate_arpc_2(sk_ac, arqc, csu)
    assert arpc.hex().upper() == "7DFB1188"



import pytest
from pyemv import cvv


def test_generate_ac_exception() -> None:
    # icc cvc3 != 16 bytes
    with pytest.raises(
        ValueError,
        match="ICC CVC3 key must be a double length DES key",
    ):
        cvv.generate_cvc3(
            bytes.fromhex("AAAAAAAAAAAAAAAA"),
            bytes.fromhex("5123456789012345D35121010000000000000F"),
            bytes.fromhex("005E"),
            bytes.fromhex("00000899"),
        )

    # ATC != 2 bytes
    with pytest.raises(
        ValueError,
        match="ATC value must be 2 bytes long",
    ):
        cvv.generate_cvc3(
            bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            bytes.fromhex("5123456789012345D35121010000000000000F"),
            bytes.fromhex("5E"),
            bytes.fromhex("00000899"),
        )

    # UN != 4 bytes
    with pytest.raises(
        ValueError,
        match="Unpredictable number must be 4 bytes long",
    ):
        cvv.generate_cvc3(
            bytes.fromhex("AAAAAAAAAAAAAAAABBBBBBBBBBBBBBBB"),
            bytes.fromhex("5123456789012345D35121010000000000000F"),
            bytes.fromhex("005E"),
            bytes.fromhex("000899"),
        )



from pyemv import cvv, kd, tools


def test_generate_cvc3_track2() -> None:
    # Device ICC CVC3 key from ISS CVC3 key using method A
    # Issuer may choose to use method A or B.
    iss_cvc3 = bytes.fromhex("01234567899876543210012345678998")
    pan = "5123456789012345"
    psn = "00"
    icc_cvc3 = kd.derive_icc_mk_a(iss_cvc3, pan, psn)
    assert tools.key_check_digits(icc_cvc3).hex().upper() == "BB8F"

    # Compute CVC3 - track2
    # Track2 must be converted to binary data where, for example,
    # characters "D3" are represented as "\xD3"
    track2 = bytes.fromhex("5123456789012345D35121010000000000000F")
    atc = bytes.fromhex("005E")
    un = bytes.fromhex("00000899")
    cvc3 = cvv.generate_cvc3(icc_cvc3, track2, atc, un)
    assert cvc3[-3:] == "488"


def test_generate_cvc3_track1() -> None:
    # Device ICC CVC3 key from ISS CVC3 key using method A
    # Issuer may choose to use method A or B.
    iss_cvc3 = bytes.fromhex("01234567899876543210012345678998")
    pan = "5123456789012345"
    psn = "00"
    icc_cvc3 = kd.derive_icc_mk_a(iss_cvc3, pan, psn)
    assert tools.key_check_digits(icc_cvc3).hex().upper() == "BB8F"

    # Compute CVC3 - track1
    # Track1 must be converted to binary data where, for example,
    # charater "B" is represented as "\x42".
    track1 = b"B5123456789012345^KENOBI/OBIWAN^3512101000000000000000000000000"
    atc = bytes.fromhex("005E")
    un = bytes.fromhex("00000899")
    cvc3 = cvv.generate_cvc3(icc_cvc3, track1, atc, un)
    assert cvc3[-3:] == "419"



"""Cryptogram Version Number (CVN) defines the following:

    - ICC Master Key derivation method
    - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method
    - Application Cryptogram (ARQC, TC, ACC) calculation method
    - Authorisation Response Cryptogram Session Key derivation method
    - Authorisation Response Cryptogram calculation method
    - Secure Messaging Session Key derivation method
    - Secure Messaging Integrity (MAC) data format and padding
    - Secure Messaging Confidentiality encryption method and padding

CVN module combines application cryptogram generation, key derivation and
secure messaging in one class per CVN.

    >>> import pyemv
    >>> cvn18 = pyemv.cvn.VisaCVN18(
    ...     iss_mk_ac=bytes.fromhex('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'),
    ...     iss_mk_smi=bytes.fromhex('BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'),
    ...     iss_mk_smc=bytes.fromhex('CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC'),
    ...     pan='1234567890123456',
    ...     psn='00')

    >>> atc = bytes.fromhex('0FFF')
    >>> arqc = cvn18.generate_ac(
    ...     tag_9f02=bytes.fromhex('000000009999'),
    ...     tag_9f03=bytes.fromhex('000000000000'),
    ...     tag_9f1a=bytes.fromhex('0840'),
    ...     tag_95=bytes.fromhex('8000048000'),
    ...     tag_5f2a=bytes.fromhex('0840'),
    ...     tag_9a=bytes.fromhex('991231'),
    ...     tag_9c=bytes.fromhex('01'),
    ...     tag_9f37=bytes.fromhex('52BF4585'),
    ...     tag_82=bytes.fromhex('1800'),
    ...     tag_9f36=atc,
    ...     tag_9f10=bytes.fromhex('06011203A0B800'))
    >>> arqc.hex().upper()
    '769577B5ABE9FE62'

    >>> arpc = cvn18.generate_arpc(
    ...     tag_9f26=arqc,
    ...     tag_9f36=atc,
    ...     csu=bytes.fromhex('00000000'))
    >>> arpc.hex().upper()
    '76503F48'

    >>> command_mac = cvn18.generate_command_mac(
    ...         command_header=bytes.fromhex('8418000008'),
    ...         tag_9f26=arqc,
    ...         tag_9f36=atc)
    >>> command_mac.hex().upper()
    'B5CB29759F9C3919'

    >>> pin_command = cvn18.generate_pin_change_command(
    ...         pin='9999',
    ...         tag_9f26=arqc,
    ...         tag_9f36=atc)
    >>> pin_command.hex().upper()
    '84240002182DC7A061323BA62472BC5308BD291B5F665B3A927E60661E'
"""


import typing as _typing

from pyemv import ac as _ac
from pyemv import kd as _kd
from pyemv import sm as _sm

__all__ = [
    "VisaCVN10",
    "VisaCVN18",
    "VisaCVN22",
    "InteracCVN133",
    "MasterCardCVN16",
    "MasterCardCVN17",
    "MasterCardCVN20",
    "MasterCardCVN21",
]


class VisaCVN10:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    Visa Cryptogram Version Number (CVN) 10 defines the following:

        - ICC Master Key derivation method = Option A
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = None
        - Application Cryptogram (ARQC, TC, ACC) calculation method = Visa
        - Authorisation Response Cryptogram Session Key derivation method = None
        - Authorisation Response Cryptogram calculation method = 1
        - Secure Messaging Session Key derivation method = Visa
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = Visa

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option A.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_a(iss_mk_smc, pan, psn)

    def _derive_sk_ac_none(self) -> bytes:
        """Derive Application Cryptogram Session Key.
        Use ICC Master Key, since Session Key is not applicable.

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return self.icc_mk_ac

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        cvr: bytes,
    ) -> bytes:
        """Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        cvr : bytes
            4 bytes of binary Card Verification Results extracted from
            9F10 (Issuer Application Data).

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_none(),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + cvr,
            _ac.PaddingType.VISA,
            8,
        )

    def generate_arpc(self, tag_9f26: bytes, arpc_rc: bytes) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 1.
        Method for the generation of a 8-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 2-byte binary ARPC response code

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        arpc_rc : bytes
            Binary 2-byte ARPC response code.

        Returns
        -------
        arpc : bytes
            Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || ARPC-RC
        """
        return _ac.generate_arpc_1(self._derive_sk_ac_none(), tag_9f26, arpc_rc)

    def _derive_sk_sm_visa(self, icc_mk_sm: bytes, tag_9f36: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Visa method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_visa_sm_sk(icc_mk_sm, tag_9f36)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.
        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_visa(self.icc_mk_smi, tag_9f36),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f36: bytes) -> bytes:
        """Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_visa(self.icc_mk_smc, tag_9f36),
            command_data,
            _sm.EncryptionType.VISA,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
        current_pin: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> bytes:
        """Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        current_pin : bytes or str, optional
            Current ASCII Personal Identification Number (optional). If present
            VIS PIN block is generated using current PIN.

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.
        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_vis_pin_block(self.icc_mk_ac, pin, current_pin), tag_9f36
        )

        if current_pin is None:
            command_header = b"\x84\x24\x00\x02\x18"
        else:
            command_header = b"\x84\x24\x00\x01\x18"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )


class VisaCVN18:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    Visa Cryptogram Version Number (CVN) 18 defines the following:

        - ICC Master Key derivation method = Option B
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = Common
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = Common
        - Authorisation Response Cryptogram calculation method = 2
        - Secure Messaging Session Key derivation method = Visa
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = Visa

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option B.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_b(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_b(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_b(iss_mk_smc, pan, psn)

    def _derive_sk_ac_common(self, tag_9f36: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using EMV Common method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 6)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        tag_9f10: bytes,
    ) -> bytes:
        """Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        tag_9f10 : bytes
            Binary data from tag 9F10 (Issuer Application Data).

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_common(tag_9f36),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + tag_9f10,
            _ac.PaddingType.EMV,
            8,
        )

    def generate_arpc(
        self,
        tag_9f26: bytes,
        tag_9f36: bytes,
        csu: bytes,
        proprietary_auth_data: _typing.Optional[bytes] = None,
    ) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 2.
        Method for the generation of a 4-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 4-byte binary Card Status Update (CSU)
            - 0-8 byte binary Proprietary Authentication Data

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        csu : bytes
            Binary 4-byte Card Status Update (CSU).
        prop_auth_data : bytes, optional
            Binary 0-8 byte Proprietary Authentication Data.

        Returns
        -------
        arpc : bytes
            Returns binary 4-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || CSU || { Proprietary Authentication Data }
        """
        return _ac.generate_arpc_2(
            self._derive_sk_ac_common(tag_9f36),
            tag_9f26,
            csu,
            proprietary_auth_data,
        )

    def _derive_sk_sm_visa(self, icc_mk_sm: bytes, tag_9f36: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Visa method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_visa_sm_sk(icc_mk_sm, tag_9f36)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_visa(self.icc_mk_smi, tag_9f36),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f36: bytes) -> bytes:
        """Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_visa(self.icc_mk_smc, tag_9f36),
            command_data,
            _sm.EncryptionType.VISA,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
        current_pin: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> bytes:
        """Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        current_pin : bytes or str, optional
            Current ASCII Personal Identification Number (optional). If present
            VIS PIN block is generated using current PIN.

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_vis_pin_block(self.icc_mk_ac, pin, current_pin), tag_9f36
        )

        if current_pin is None:
            command_header = b"\x84\x24\x00\x02\x18"
        else:
            command_header = b"\x84\x24\x00\x01\x18"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )


class VisaCVN22:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    Visa Cryptogram Version Number (CVN) 18 defines the following:

        - ICC Master Key derivation method = Option B
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = Common
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = Common
        - Authorisation Response Cryptogram calculation method = 2
        - Secure Messaging Session Key derivation method = Common
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = Visa

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option B.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_b(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_b(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_b(iss_mk_smc, pan, psn)

    def _derive_sk_ac_common(self, tag_9f36: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using EMV Common method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 6)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        tag_9f10: bytes,
    ) -> bytes:
        """Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        tag_9f10 : bytes
            Binary data from tag 9F10 (Issuer Application Data).

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_common(tag_9f36),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + tag_9f10,
            _ac.PaddingType.EMV,
            8,
        )

    def generate_arpc(
        self,
        tag_9f26: bytes,
        tag_9f36: bytes,
        csu: bytes,
        proprietary_auth_data: _typing.Optional[bytes] = None,
    ) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 2.
        Method for the generation of a 4-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 4-byte binary Card Status Update (CSU)
            - 0-8 byte binary Proprietary Authentication Data

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        csu : bytes
            Binary 4-byte Card Status Update (CSU).
        prop_auth_data : bytes, optional
            Binary 0-8 byte Proprietary Authentication Data.

        Returns
        -------
        arpc : bytes
            Returns binary 4-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || CSU || { Proprietary Authentication Data }
        """
        return _ac.generate_arpc_2(
            self._derive_sk_ac_common(tag_9f36),
            tag_9f26,
            csu,
            proprietary_auth_data,
        )

    def _derive_sk_sm_common(self, icc_mk_sm: bytes, tag_9f26: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Common method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(icc_mk_sm, tag_9f26)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_common(self.icc_mk_smi, tag_9f26),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f26: bytes) -> bytes:
        """Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_common(self.icc_mk_smc, tag_9f26),
            command_data,
            _sm.EncryptionType.VISA,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
        current_pin: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> bytes:
        """Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        current_pin : bytes or str, optional
            Current ASCII Personal Identification Number (optional). If present
            VIS PIN block is generated using current PIN.

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        Notes
        -----
        This command generates Visa PIN block format when current PIN is
        supplied and ISO 9564-1 format 2 PIN block when it's not supplied.
        Some issuers may choose to use ISO 9564-1 format 1 PIN block when
        current PIN is not supplied.
        """

        if current_pin is None:
            # Use standard EMV PIN block format if current PIN is not required
            command_header = b"\x84\x24\x00\x02\x18"
            enc_pin_block = self.encrypt_command_data(
                _sm.format_iso9564_2_pin_block(pin), tag_9f26
            )
        else:
            # Use legacy Visa PIN block format if current PIN is required
            command_header = b"\x84\x24\x00\x01\x18"
            enc_pin_block = self.encrypt_command_data(
                _sm.format_vis_pin_block(self.icc_mk_ac, pin, current_pin), tag_9f26
            )

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )


class InteracCVN133:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    Interac Cryptogram Version Number (CVN) 133 defines the following:

        - ICC Master Key derivation method = Option A
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = MasterCard
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = MasterCard
        - Authorisation Response Cryptogram calculation method = 1
        - Secure Messaging Session Key derivation method = MasterCard
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          not padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = MasterCard

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option A.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_a(iss_mk_smc, pan, psn)

    def _derive_sk_ac_mastercard(self, tag_9f36: bytes, tag_9f37: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using MasterCard method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 2 + tag_9f37)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        tag_9f10: bytes,
    ) -> bytes:
        """Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        tag_9f10 : bytes
            Binary data from tag 9F10 (Issuer Application Data).

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_mastercard(tag_9f36, tag_9f37),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + tag_9f10,
            _ac.PaddingType.EMV,
            8,
        )

    def generate_arpc(
        self,
        tag_9f26: bytes,
        tag_9f37: bytes,
        tag_9f36: bytes,
        arpc_rc: bytes,
    ) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 1.
        Method for the generation of a 8-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 2-byte binary ARPC response code

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        arpc_rc : bytes
            Binary 2-byte ARPC response code.

        Returns
        -------
        arpc : bytes
            Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || ARPC-RC
        """
        return _ac.generate_arpc_1(
            self._derive_sk_ac_mastercard(tag_9f36, tag_9f37),
            tag_9f26,
            arpc_rc,
        )

    def _derive_sk_sm_common(self, icc_mk_sm: bytes, tag_9f26: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Common method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(icc_mk_sm, tag_9f26)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_common(self.icc_mk_smi, tag_9f26),
            command_header + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f26: bytes) -> bytes:
        """Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_common(self.icc_mk_smc, tag_9f26),
            command_data,
            _sm.EncryptionType.MASTERCARD,
        )

    def generate_pin_change_command(
        self, pin: _typing.Union[bytes, str], tag_9f26: bytes
    ) -> bytes:
        """Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_iso9564_2_pin_block(pin), tag_9f26
        )

        command_header = b"\x84\x24\x00\x02\x10"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(command_header, tag_9f26, enc_pin_block)
        )


class MasterCardCVN16:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    MasterCard Cryptogram Version Number (CVN) 16 defines the following:

        - ICC Master Key derivation method = Option A
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = MasterCard
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = None
        - Authorisation Response Cryptogram calculation method = 1
        - Secure Messaging Session Key derivation method = Common
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = MasterCard

    MasterCard CVN 16 (binary 0001 0 00 0):

        - Bit 2-3 = 00: Uses Mastercard Proprietary SKD session key
        - Bit 1 = 0:    Does not include counters in AC generation

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option A.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_a(iss_mk_smc, pan, psn)

    def _derive_sk_ac_mastercard(self, tag_9f36: bytes, tag_9f37: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using MasterCard method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 2 + tag_9f37)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        cvr: bytes,
    ) -> bytes:
        r"""Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        cvr : bytes
            6 bytes of binary Card Verification Results extracted from
            9F10 (Issuer Application Data).

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_mastercard(tag_9f36, tag_9f37),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + cvr,
            _ac.PaddingType.EMV,
            8,
        )

    def _derive_sk_arpc_none(self) -> bytes:
        """Derive Application Cryptogram Session Key.
        Use ICC Master Key, since Session Key is not applicable.

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return self.icc_mk_ac

    def generate_arpc(self, tag_9f26: bytes, arpc_rc: bytes) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 1.
        Method for the generation of a 8-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 2-byte binary ARPC response code

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        arpc_rc : bytes
            Binary 2-byte ARPC response code.

        Returns
        -------
        arpc : bytes
            Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || ARPC-RC
        """
        return _ac.generate_arpc_1(self._derive_sk_arpc_none(), tag_9f26, arpc_rc)

    def _derive_sk_sm_common(self, icc_mk_sm: bytes, tag_9f26: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Common method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(icc_mk_sm, tag_9f26)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_common(self.icc_mk_smi, tag_9f26),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f26: bytes) -> bytes:
        r"""Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_common(self.icc_mk_smc, tag_9f26),
            command_data,
            _sm.EncryptionType.MASTERCARD,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
    ) -> bytes:
        r"""Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_iso9564_2_pin_block(pin), tag_9f26
        )

        command_header = b"\x84\x24\x00\x02\x10"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )


class MasterCardCVN17:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    MasterCard Cryptogram Version Number (CVN) 17 defines the following:

        - ICC Master Key derivation method = Option A
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = MasterCard
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = None
        - Authorisation Response Cryptogram calculation method = 1
        - Secure Messaging Session Key derivation method = Common
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = MasterCard

    MasterCard CVN 17 (binary 0001 0 00 1):

        - Bit 2-3 = 00: Uses Mastercard Proprietary SKD session key
        - Bit 1 = 1:    Includes counters in AC generation

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option A.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_a(iss_mk_smc, pan, psn)

    def _derive_sk_ac_mastercard(self, tag_9f36: bytes, tag_9f37: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using MasterCard method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 2 + tag_9f37)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        cvr: bytes,
        counters: bytes,
    ) -> bytes:
        r"""Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        cvr : bytes
            6 bytes of binary Card Verification Results extracted from
            9F10 (Issuer Application Data).
        counters : bytes
            Counters include Cumulative Offline Transaction Amount
            (6 bytes), Consecutive Offline Transactions Number (1 byte) and
            1 byte set to \xFF.
            It's assumed that the counters are in the clear; not encrypted.
            These fields are extracted from 9F10 (Issuer Application Data).
            For M/Chip 4 counters are always 8 bytes long.
            For M/Chip Advance counters can be 8 or 16 bytes, with or without
            last online ATC.

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_mastercard(tag_9f36, tag_9f37),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + cvr
            + counters,
            _ac.PaddingType.EMV,
            8,
        )

    def _derive_sk_arpc_none(self) -> bytes:
        """Derive Application Cryptogram Session Key.
        Use ICC Master Key, since Session Key is not applicable.

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return self.icc_mk_ac

    def generate_arpc(self, tag_9f26: bytes, arpc_rc: bytes) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 1.
        Method for the generation of a 8-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 2-byte binary ARPC response code

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        arpc_rc : bytes
            Binary 2-byte ARPC response code.

        Returns
        -------
        arpc : bytes
            Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || ARPC-RC
        """
        return _ac.generate_arpc_1(self._derive_sk_arpc_none(), tag_9f26, arpc_rc)

    def _derive_sk_sm_common(self, icc_mk_sm: bytes, tag_9f26: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Common method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(icc_mk_sm, tag_9f26)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_common(self.icc_mk_smi, tag_9f26),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f26: bytes) -> bytes:
        r"""Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_common(self.icc_mk_smc, tag_9f26),
            command_data,
            _sm.EncryptionType.MASTERCARD,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
    ) -> bytes:
        r"""Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_iso9564_2_pin_block(pin), tag_9f26
        )

        command_header = b"\x84\x24\x00\x02\x10"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )


class MasterCardCVN20:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    MasterCard Cryptogram Version Number (CVN) 20 defines the following:

        - ICC Master Key derivation method = Option A
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = Common
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = Common
        - Authorisation Response Cryptogram calculation method = 1
        - Secure Messaging Session Key derivation method = Common
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = MasterCard

    MasterCard CVN 20 (binary 0001 0 10 0):

        - Bit 2-3 = 10: Uses EMV CSK session key
        - Bit 1 = 0:    Does not include counters in AC generation

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option A.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_a(iss_mk_smc, pan, psn)

    def _derive_sk_ac_common(self, tag_9f36: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using EMV Common method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 6)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        cvr: bytes,
    ) -> bytes:
        r"""Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        cvr : bytes
            6 bytes of binary Card Verification Results extracted from
            9F10 (Issuer Application Data).

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_common(tag_9f36),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + cvr,
            _ac.PaddingType.EMV,
            8,
        )

    def generate_arpc(self, tag_9f26: bytes, tag_9f36: bytes, arpc_rc: bytes) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 1.
        Method for the generation of a 8-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 2-byte binary ARPC response code

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        arpc_rc : bytes
            Binary 2-byte ARPC response code.

        Returns
        -------
        arpc : bytes
            Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || ARPC-RC
        """
        return _ac.generate_arpc_1(
            self._derive_sk_ac_common(tag_9f36), tag_9f26, arpc_rc
        )

    def _derive_sk_sm_common(self, icc_mk_sm: bytes, tag_9f26: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Common method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(icc_mk_sm, tag_9f26)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_common(self.icc_mk_smi, tag_9f26),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f26: bytes) -> bytes:
        r"""Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_common(self.icc_mk_smc, tag_9f26),
            command_data,
            _sm.EncryptionType.MASTERCARD,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
    ) -> bytes:
        r"""Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_iso9564_2_pin_block(pin), tag_9f26
        )

        command_header = b"\x84\x24\x00\x02\x10"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )


class MasterCardCVN21:
    """Cryptogram Version Number (CVN) defines Card Authentication Method (CAM).
    MasterCard Cryptogram Version Number (CVN) 21 defines the following:

        - ICC Master Key derivation method = Option A
        - Application Cryptogram (ARQC, TC, ACC) Session Key derivation method = Common
        - Application Cryptogram (ARQC, TC, ACC) calculation method = EMV
        - Authorisation Response Cryptogram Session Key derivation method = Common
        - Authorisation Response Cryptogram calculation method = 1
        - Secure Messaging Session Key derivation method = Common
        - Secure Messaging Integrity (MAC) data format and padding = Format 2,
          padded with transaction data
        - Secure Messaging Confidentiality encryption method and padding = MasterCard

    MasterCard CVN 20 (binary 0001 0 10 1):

        - Bit 2-3 = 10: Uses EMV CSK session key
        - Bit 1 = 1:    Includes counters in AC generation

    Parameters
    ----------
    iss_mk_ac : bytes
        16-byte binary Issuer Master Key for Application Cryptography.
        Has to be a valid DES key.
    iss_mk_smi : bytes
        16-byte binary Issuer Master Key for Issuer Script Integrity.
        Has to be a valid DES key.
    iss_mk_smc : bytes
        16-byte binary Issuer Master Key for Issuer Script Confidentiality.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Attributes
    ----------
    icc_mk_ac : bytes
        16-byte binary ICC Master Key for Application Cryptography.
    icc_mk_smi : bytes
        16-byte binary ICC Master Key for Issuer Script Integrity.
    icc_mk_smc : bytes
        16-byte binary ICC Master Key for Issuer Script Confidentiality.
    """

    def __init__(
        self,
        iss_mk_ac: bytes,
        iss_mk_smi: bytes,
        iss_mk_smc: bytes,
        pan: _typing.Union[bytes, str],
        psn: _typing.Optional[_typing.Union[bytes, str]] = None,
    ) -> None:
        # Derive AC, SMI, and SMC ICC Master Keys for a new card
        # using option A.
        psn = psn or "00"
        self.icc_mk_ac = _kd.derive_icc_mk_a(iss_mk_ac, pan, psn)
        self.icc_mk_smi = _kd.derive_icc_mk_a(iss_mk_smi, pan, psn)
        self.icc_mk_smc = _kd.derive_icc_mk_a(iss_mk_smc, pan, psn)

    def _derive_sk_ac_common(self, tag_9f36: bytes) -> bytes:
        """Derive Application Cryptogram Session Key using EMV Common method.

        Parameters
        ----------
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        sk_ac : bytes
            16-byte binary Session Key for Application Cryptogram.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(self.icc_mk_ac, tag_9f36 + b"\x00" * 6)

    def generate_ac(
        self,
        tag_9f02: bytes,
        tag_9f03: bytes,
        tag_9f1a: bytes,
        tag_95: bytes,
        tag_5f2a: bytes,
        tag_9a: bytes,
        tag_9c: bytes,
        tag_9f37: bytes,
        tag_82: bytes,
        tag_9f36: bytes,
        cvr: bytes,
        counters: bytes,
    ) -> bytes:
        r"""Generate Application Cryptogram. Same process for
            - Authorisation Request Cryptogram (ARQC)
            - Transaction Cryptogram (TC)
            - Application Authentication Cryptogram (AAC)

        Parameters
        ----------
        tag_9f02 : bytes
            Binary data from tag 9F02 (Amount, Authorized).
        tag_9f03 : bytes
            Binary data from tag 9F03 (Amount, Other).
        tag_9f1a : bytes
            Binary data from tag 9F1A (Terminal Country Code).
        tag_95 : bytes
            Binary data from tag 95 (Terminal Verification Results).
        tag_5f2a : bytes
            Binary data from tag 5F2A (Transaction Currency Code).
        tag_9a : bytes
            Binary data from tag 9A (Transaction Date).
        tag_9c : bytes
            Binary Data from tag 9C (Transaction Type).
        tag_9f37 : bytes
            Binary data from tag 9F37 (Unpredictable Number).
        tag_82 : bytes
            Binary data from tag 82 (Application Interchange Profile).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        cvr : bytes
            6 bytes of binary Card Verification Results extracted from
            9F10 (Issuer Application Data).
        counters : bytes
            Counters include Cumulative Offline Transaction Amount
            (6 bytes), Consecutive Offline Transactions Number (1 byte) and
            1 byte set to \xFF.
            It's assumed that the counters are in the clear; not encrypted.
            These fields are extracted from 9F10 (Issuer Application Data).
            For M/Chip 4 counters are always 8 bytes long.
            For M/Chip Advance counters can be 8 or 16 bytes, with or without
            last online ATC.

        Returns
        -------
        ac : bytes
            Returns binary 8-byte cryptogram (ARQC, TC, AAC).
        """
        return _ac.generate_ac(
            self._derive_sk_ac_common(tag_9f36),
            tag_9f02
            + tag_9f03
            + tag_9f1a
            + tag_95
            + tag_5f2a
            + tag_9a
            + tag_9c
            + tag_9f37
            + tag_82
            + tag_9f36
            + cvr
            + counters,
            _ac.PaddingType.EMV,
            8,
        )

    def generate_arpc(self, tag_9f26: bytes, tag_9f36: bytes, arpc_rc: bytes) -> bytes:
        """Generate Authorisation Response Cryptogram (ARPC) using method 1.
        Method for the generation of a 8-byte ARPC consists of applying
        ISO/IEC 9797-1 MAC algorithm 3 to:

            - 8-byte binary ARQC
            - 2-byte binary ARPC response code

        Parameters
        ----------
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        arpc_rc : bytes
            Binary 2-byte ARPC response code.

        Returns
        -------
        arpc : bytes
            Returns binary 8-byte Authorisation Response Cryptogram (ARPC).
            The resulting issuer authentication data (tag 91) is:

                91 || Len || ARPC || ARPC-RC
        """
        return _ac.generate_arpc_1(
            self._derive_sk_ac_common(tag_9f36), tag_9f26, arpc_rc
        )

    def _derive_sk_sm_common(self, icc_mk_sm: bytes, tag_9f26: bytes) -> bytes:
        """Derive Secure Messaging Session Key using Common method.

        Parameters
        ----------
        icc_mk_sm : bytes
            16-byte binary ICC Master Key for Secure Messaging.
            Has to be a valid DES key.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).

        Returns
        -------
        sk_sm : bytes
            16-byte binary Session Key for Secure Messaging.
            Has to be a valid DES key.
        """
        return _kd.derive_common_sk(icc_mk_sm, tag_9f26)

    def generate_command_mac(
        self,
        command_header: bytes,
        tag_9f26: bytes,
        tag_9f36: bytes,
        command_data: bytes = b"",
    ) -> bytes:
        r"""Message Authentication Code (MAC) for Secure Messaging Integrity.

        Parameters
        ----------
        command_header : bytes
            Binary command header, such as \x84\x24\x00\x00\x08 for PIN unblock.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).
        command_data : bytes, optional
            Binary command data, e.g. PUT DATA or PIN block.

        Returns
        -------
        mac : bytes
            Binary 8-byte command MAC.

        """
        return _sm.generate_command_mac(
            self._derive_sk_sm_common(self.icc_mk_smi, tag_9f26),
            command_header + tag_9f36 + tag_9f26 + command_data,
            8,
        )

    def encrypt_command_data(self, command_data: bytes, tag_9f26: bytes) -> bytes:
        r"""Command Data Encryption for Secure Messaging Confidentiality.

        Parameters
        ----------
        command_data : bytes
            Binary command data, e.g. PUT DATA or PIN block.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.

        Returns
        -------
        encrypted_command_data : bytes
            Binary encrypted command command data. Then the resulting command is:

                Header || Encrypted Data || MAC
        """
        return _sm.encrypt_command_data(
            self._derive_sk_sm_common(self.icc_mk_smc, tag_9f26),
            command_data,
            _sm.EncryptionType.MASTERCARD,
        )

    def generate_pin_change_command(
        self,
        pin: _typing.Union[bytes, str],
        tag_9f26: bytes,
        tag_9f36: bytes,
    ) -> bytes:
        r"""Generate a PIN change command with encrypted PIN block and MAC.

        Parameters
        ----------
        pin : bytes or str
            New ASCII Personal Identification Number.
        tag_9f26 : bytes
            Binary data from tag 9F26 (Authorisation Request Cryptogram).
            This value should be increment by 1 for each process script command
            after the first command. For example, the first script command will
            use 9F26 as-is (e.g. \x12\x34\x56\x78\x12\x34\x56\x78).
            The second script command will use 9F26 + 1
            (e.g. \x12\x34\x56\x78\x12\x34\x56\x79). And so on.
        tag_9f36 : bytes
            Binary data from tag 9F36 (Application Transaction Counter).

        Returns
        -------
        pin_change_command : bytes
            Binary PIN change command with encrypted PIN block and MAC.

        """
        enc_pin_block = self.encrypt_command_data(
            _sm.format_iso9564_2_pin_block(pin), tag_9f26
        )

        command_header = b"\x84\x24\x00\x02\x10"

        return (
            command_header
            + enc_pin_block
            + self.generate_command_mac(
                command_header, tag_9f26, tag_9f36, enc_pin_block
            )
        )



r"""Cryptographic key derivation functions for the ICC Master Keys and ICC Session Keys.

ICC Master Key derivation method A:

    >>> import pyemv
    >>> iss_mk = bytes.fromhex('0123456789ABCDEFFEDCBA9876543210')
    >>> pan = '99012345678901234'
    >>> psn = '45'
    >>> icc_mk = pyemv.kd.derive_icc_mk_a(iss_mk, pan, psn)
    >>> icc_mk.hex().upper()
    '67F8292358083E5EA7AB7FDA58D53B6B'
"""

import binascii as _binascii
import hashlib as _hashlib
import typing as _typing

from pyemv import tools as _tools

__all__ = [
    "derive_icc_mk_a",
    "derive_icc_mk_b",
    "derive_common_sk",
    "derive_visa_sm_sk",
    "derive_emv2000_tree_sk",
]


def derive_icc_mk_a(
    iss_mk: bytes,
    pan: _typing.Union[bytes, str],
    psn: _typing.Optional[_typing.Union[bytes, str]] = None,
) -> bytes:
    r"""ICC Master Key Derivation. EMV Option A.
    Uses PAN, PAN Sequence Number, MK ISS, Triple DES.

    Parameters
    ----------
    iss_mk : bytes
        Binary Issuer Master Key to derive ICC Master Key from.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Returns
    -------
    icc_mk : bytes
        Binary 16-byte ICC Master Key.

    Notes
    -----
    Derived from Issuer Master Key (iss_mk).
    Uses EMV Option A - Master Key Derivation method which uses
    the PAN and PAN sequence number, as defined in EMV Book 2, Annex A.

    When a card is personalised the issuer will take the 3 iss_mk keys
    and calculate the 3 icc_mk keys to be stored on the card.

        - icc_mk_ac - used for the transaction cryptograms (ARQC, TC or AAC)
        - icc_mk_smi - used for Issuer Script Integrity
        - icc_mk_smc - used for Issuer Script Confidentiality

    For further details see also:
        - EMV 4.3 Book 2 Annex A 1.4 Master Key Derivation
        - EMV 4.3 Book 2 Annex A 1.4.1 Option A

    Examples
    --------
    >>> from pyemv import kd
    >>> iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> icc_mk = kd.derive_icc_mk_a(iss_mk, pan="12345678901234567", psn="01")
    >>> icc_mk.hex().upper()
    '73AD54688CEF2934B0979857E3C719F1'
    """
    if psn is None:
        psn = "00"

    if isinstance(psn, bytes):
        psn = psn.decode("ascii")

    if isinstance(pan, bytes):
        pan = pan.decode("ascii")

    # Data A must be at most 16 digits, right-justified,
    # zero-padded from the left.
    data_a = _binascii.a2b_hex((pan + psn)[-16:].zfill(16))

    # Data B is inverted data A
    data_b = _tools.xor(data_a, b"\xFF" * len(data_a))

    icc_mk = _tools.encrypt_tdes_ecb(iss_mk, data_a + data_b)

    return _tools.adjust_key_parity(icc_mk)


def derive_icc_mk_b(
    iss_mk: bytes,
    pan: _typing.Union[bytes, str],
    psn: _typing.Optional[_typing.Union[bytes, str]] = None,
) -> bytes:
    r"""ICC Master Key Derivation. EMV Option B.
    Uses PAN, PAN Sequence Number, MK ISS, Triple DES, SHA-1 and
    decimalisation of hex digits.

    Parameters
    ----------
    iss_mk : bytes
        Binary Issuer Master Key to derive ICC Master Key from.
        Has to be a valid DES key.
    pan : bytes or str
        ASCII Application Primary Account Number.
    psn : bytes or str, optional
        ASCII 2-digit PAN Sequence Number (default 00).

    Returns
    -------
    icc_mk : bytes
        Binary 16-byte ICC Master Key

    Notes
    -----
    Derived from Issuer Master Key (iss_mk).
    Uses EMV Option B - Master Key Derivation method which uses
    the PAN and PAN sequence number, as defined in EMV Book 2, Annex A.

    When a card is personalised the issuer will take the 3 iss_mk keys
    and calculate the 3 icc_mk keys to be stored on the card.

        - icc_mk_ac - used for the transaction cryptograms (ARQC, TC or AAC)
        - icc_mk_smi - used for Issuer Script Integrity
        - icc_mk_smc - used for Issuer Script Confidentiality

    For further details see also:
        - EMV 4.3 Book 2 Annex A 1.4 Master Key Derivation
        - EMV 4.3 Book 2 Annex A 1.4.2 Option B

    Examples
    --------
    >>> from pyemv import kd
    >>> iss_mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> icc_mk = kd.derive_icc_mk_b(iss_mk, pan="12345678901234567", psn="01")
    >>> icc_mk.hex().upper()
    'AD406D7F6D7570916D75E5DCAB8CF737'
    """
    # For PANs with length of 16 or less method B works as method A
    if len(pan) <= 16:
        return derive_icc_mk_a(iss_mk, pan, psn)

    if psn is None:
        psn = "00"

    if isinstance(psn, bytes):
        psn = psn.decode("ascii")

    if isinstance(pan, bytes):
        pan = pan.decode("ascii")

    # Data A must be an even number of digits,
    # right-justified, zero-padded from the left.
    if len(pan) % 2:
        pan_psn = _binascii.a2b_hex("0" + pan + psn)
    else:
        pan_psn = _binascii.a2b_hex(pan + psn)

    # Hash PAN || PAN sequence
    digest = _hashlib.sha1(pan_psn).hexdigest()

    # Get first 16 digits out the hash value.
    result = "".join(
        [d for d in digest if d in {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"}][
            :16
        ]
    )

    # If there are not enough digits, substitute
    # letters using the following decimalization table:
    # Input a b c d e f
    # Table 0 1 2 3 4 5
    if len(result) < 16:
        digest = "".join(
            [d for d in digest if d in {"a", "b", "c", "d", "e", "f"}][
                : 16 - len(result)
            ]
        )
        digest = digest.translate({97: 48, 98: 49, 99: 50, 100: 51, 101: 52, 102: 53})
        result = result + digest

    data_a = _binascii.a2b_hex(result)

    # Data B is inverted data A
    data_b = _tools.xor(data_a, b"\xFF" * len(data_a))

    icc_mk = _tools.encrypt_tdes_ecb(iss_mk, data_a + data_b)

    return _tools.adjust_key_parity(icc_mk)


def derive_common_sk(icc_mk: bytes, r: _typing.Union[bytes, bytearray]) -> bytes:
    r"""EMV Common Session Key Derivation.

    Parameters
    ----------
    icc_mk : bytes
        Binary ICC Master Key to derive session key from.
        Has to be a valid DES key.
    r : bytes, bytearray
        Binary diversification value. Examples of diversification value:

            - R = ATC || 00 || 00 || 00 || 00 || 00 || 00 (AC Session Keys)
            - R = ARQC (Secure Messaging for Integrity and Confidentiality
              Session Keys)
            - R = ATC || 00 || 00 || UN (AC Session Keys)
            - Any other proprietary value

    Returns
    -------
    sk : bytes
        Binary 16-byte Session Key.

    Raises
    ------
    ValueError
        ICC Master Key must be a double length DES key
    ValueError
        Diversification value must be 8 bytes long

    Notes
    -----
    For more information see:
        - EMV 4.3 Book 2 Annex A 1.3 Session Key Derivation
        - EMV 4.3 Book 2 Annex A 1.3.1 Common Session Key Derivation Option

    Examples
    --------
    >>> from pyemv import kd
    >>> mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> r = bytes.fromhex("001C000000000000")
    >>> sk = kd.derive_common_sk(mk, r)
    >>> sk.hex().upper()
    'E9FB384AF807B940FEDCEA613461B0C4'
    """
    if len(icc_mk) != 16:
        raise ValueError("ICC Master Key must be a double length DES key")

    if len(r) != 8:
        raise ValueError("Diversification value must be 8 bytes long")

    # SK Key A (i.e. first 8 bytes) = TDES(icc_mk)[r]
    r_a = bytearray(r)
    r_a[2] = 0xF0

    # SK Key B (i.e. second 8 bytes) = TDES(icc_mk)[r]
    r_b = bytearray(r)
    r_b[2] = 0x0F

    sk = _tools.encrypt_tdes_ecb(icc_mk, r_a + r_b)

    return _tools.adjust_key_parity(sk)


def derive_visa_sm_sk(icc_mk: bytes, atc: bytes) -> bytes:
    r"""Visa Secure Messaging Session Key Derivation.

    Parameters
    ----------
    icc_mk : bytes
        Binary ICC Master Key to derive session key from.
        Has to be a valid DES key.
    atc : bytes
        Binary data from tag 9F36 (Application Transaction Counter).

    Returns
    -------
    sk : bytes
        Binary 16-byte Session Key.

    Raises
    ------
    ValueError
        ICC Master Key must be a double length DES key
    ValueError
        ATC value must be 2 bytes long

    Examples
    --------
    >>> from pyemv import kd
    >>> mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> atc = bytes.fromhex("001C")
    >>> sk = kd.derive_visa_sm_sk(mk, atc)
    >>> sk.hex().upper()
    '0123456789ABCDF2FEDCBA987654CDF2'
    """
    if len(icc_mk) != 16:
        raise ValueError("ICC Master Key must be a double length DES key")

    if len(atc) != 2:
        raise ValueError("ATC value must be 2 bytes long")

    # SK Key A (i.e. first 8 bytes) = r xor MK Key A
    r = b"\x00" * 6 + atc
    sk_a = _tools.xor(r, icc_mk[:8])

    # SK Key B (i.e. second 8 bytes) = r xor MK Key B
    r = b"\x00" * 6 + _tools.xor(atc, b"\xff\xff")
    sk_b = _tools.xor(r, icc_mk[8:])

    return _tools.adjust_key_parity(sk_a + sk_b)


def derive_emv2000_tree_sk(
    icc_mk: bytes,
    atc: bytes,
    height: int = 8,
    branch_factor: int = 4,
    iv: bytes = b"\x00" * 16,
) -> bytes:
    r"""EMV2000-Tree Session Key Derivation.

    Parameters
    ----------
    icc_mk : bytes
        Binary ICC Master Key to derive session key from.
        Has to be a valid DES key.
    atc : bytes
        Binary data from tag 9F36 (Application Transaction Counter).
    height : int
        Height value used for EMV-Tree derivation. Height controls
        the number of levels of intermediate keys in the tree
        excluding the base level. Set to either 8 or 16.
        The specification recommends value 8. Defaults to 8.
    branch_factor : int
        Branch factor value used for EMV-Tree derivation. Branch factor
        controls number of "child" keys a "parent" key derives.
        The specification recommends value 4. Defaults to 4.
    iv : bytes
        16-byte binary initialization vector used for EMV-Tree derivation.
        The specification recommends IV value of zeros. Defaults to 0s.

    Returns
    -------
    sk : bytes
        Binary 16-byte Session Key.

    Raises
    ------
    ValueError
        ICC Master Key must be a double length DES key
    ValueError
        ATC value must be 2 bytes long
    ValueError
        Initialization vector value must be 16 bytes long
    ValueError
        Number of possible session keys must exceed maximum ATC value

    Notes
    -----
    For more information see:
        - EMV 4.1 Book 2 Annex A 1.3 Session Key Derivation
        - EMV 4.1 Book 2 Annex A 1.3.1 Description
        - EMV 4.1 Book 2 Annex A 1.3.2 Implementation

    This method was replaced by common session key derivation in 2005
    and should not be used for new development.
    See EMVCo specification update bulletin 46 (SU-46).

    Recommended branch factor and tree height combinations are as follow.
    Both combinations produce enough session keys for every possible ATC value.
        - Branch factor 2 and tree height 16
        - Branch factor 4 and tree height 8

    Examples
    --------
    >>> from pyemv import kd
    >>> mk = bytes.fromhex("0123456789ABCDEFFEDCBA9876543210")
    >>> atc = bytes.fromhex("001C")
    >>> sk = kd.derive_emv2000_tree_sk(mk, atc, 8, 4)
    >>> sk.hex().upper()
    'E5BF6D1067F194B0A89B7F5D83BC64A2'
    """
    if len(icc_mk) != 16:
        raise ValueError("ICC Master Key must be a double length DES key")

    if len(atc) != 2:
        raise ValueError("ATC value must be 2 bytes long")

    if len(iv) != 16:
        raise ValueError("Initialization vector value must be 16 bytes long")

    # The number of possible session keys (branch_factor ** height)
    # must exceed the maximum value of the ATC which is 2 ** 16 - 1.
    if branch_factor ** height < 65535:
        raise ValueError(
            "Number of possible session keys must exceed maximum ATC value"
        )

    # F(X,Y,j) := (DES3(X)[YL XOR (j mod b)] || DES3(X)[YR XOR (j mod b) XOR 'F0'])
    def derive(x: bytes, y: bytes, j: int) -> bytes:
        """Map two 16-byte numbers X and Y and an integer j onto a 16-byte number."""
        j_mod_b = int.to_bytes(j % branch_factor, 8, "big")

        # (DES3(X)[YL XOR (j mod b)]
        l_data = _tools.xor(y[:8], j_mod_b)
        l_data = _tools.encrypt_tdes_ecb(x, l_data)

        # DES3(X)[YR XOR (j mod b) XOR 'F0']
        r_data = _tools.xor(y[8:], j_mod_b)
        r_data = _tools.xor(r_data, b"\x00" * 7 + b"\xF0")
        r_data = _tools.encrypt_tdes_ecb(x, r_data)
        return l_data + r_data

    # GP = Grandparent Key
    # IK = Intermediate Key
    # P  = Parent Key
    # H  = Height of the tree
    def walk(j: int, h: int) -> _typing.Tuple[bytes, bytes]:
        """Returns P and GP"""
        # Base case: P = ICC MK, GP = IV
        if h == 0:
            return icc_mk, iv
        p, gp = walk(j // branch_factor, h - 1)
        # Derives an IK from P and GP
        # IK becomes the new parent and current P becomes the new GP
        return derive(p, gp, j), p

    atc_num = int.from_bytes(atc, "big")

    # Derive IKs from the bottom of the tree to the second to last level
    # because GP from that level is required for SK.
    p, gp = walk(atc_num // branch_factor, height - 1)

    # Derive SK from a new IK at the tree height XOR'd by GP.
    sk = _tools.xor(derive(p, gp, atc_num), gp)
    return _tools.adjust_key_parity(sk)



from typing import Union

import pytest
from pyemv import cvn
from pyemv.tools import key_check_digits


@pytest.mark.parametrize(
    ["pan", "psn", "pin", "current_pin"],
    [
        (b"1234567890123456", b"00", b"9999", b"8888"),
        (b"1234567890123456", "00", b"9999", "8888"),
        ("1234567890123456", b"00", "9999", b"8888"),
        ("1234567890123456", "00", "9999", "8888"),
    ],
)
def test_visa_cvn10(
    pan: Union[bytes, str],
    psn: Union[bytes, str],
    pin: Union[bytes, str],
    current_pin: Union[bytes, str],
) -> None:
    """Visa CVN 10"""
    cvn10 = cvn.VisaCVN10(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Check ICC Master Keys
    assert key_check_digits(cvn10.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn10.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn10.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    arqc = cvn10.generate_ac(
        tag_9f02=bytes.fromhex("000000004000"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000048000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("191105"),
        tag_9c=bytes.fromhex("01"),
        tag_9f37=bytes.fromhex("52BF4585"),
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        cvr=bytes.fromhex("03A06010"),
    )
    assert arqc.hex().upper() == "29CCA15AE665FA2E"
    assert key_check_digits(cvn10._derive_sk_ac_none()).hex().upper() == "BAB0"

    # ARPC and its session key
    arpc = cvn10.generate_arpc(tag_9f26=arqc, arpc_rc=bytes.fromhex("3030"))
    assert arpc.hex().upper() == "28993816AFAE4AEB"
    assert key_check_digits(cvn10._derive_sk_ac_none()).hex().upper() == "BAB0"

    # MAC application unblock command and its session key
    mac = cvn10.generate_command_mac(
        bytes.fromhex("8418000008"),
        tag_9f26=arqc,
        tag_9f36=atc,
    )
    assert mac.hex().upper() == "DB56BA60087CEFD3"
    assert (
        key_check_digits(cvn10._derive_sk_sm_visa(cvn10.icc_mk_smi, atc)).hex().upper()
        == "DB10"
    )

    # PIN change without current PIN and its session keys
    pin_command = cvn10.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert (
        pin_command.hex().upper()
        == "8424000218BB221AF527377A2811D2B6EBC396A9391A965A5CB2CE62DA"
    )
    assert (
        key_check_digits(cvn10._derive_sk_sm_visa(cvn10.icc_mk_smc, atc)).hex().upper()
        == "6D04"
    )
    assert (
        key_check_digits(cvn10._derive_sk_sm_visa(cvn10.icc_mk_smi, atc)).hex().upper()
        == "DB10"
    )

    # PIN change with current PIN and its session keys
    pin_command = cvn10.generate_pin_change_command(
        pin, tag_9f26=arqc, tag_9f36=atc, current_pin=current_pin
    )
    assert (
        pin_command.hex().upper()
        == "84240001182EE025E554BCD87E11D2B6EBC396A939C4A8393B57DD2F3F"
    )
    assert (
        key_check_digits(cvn10._derive_sk_sm_visa(cvn10.icc_mk_smc, atc)).hex().upper()
        == "6D04"
    )
    assert (
        key_check_digits(cvn10._derive_sk_sm_visa(cvn10.icc_mk_smi, atc)).hex().upper()
        == "DB10"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin", "current_pin"],
    [
        (b"1234567890123456", b"00", b"9999", b"8888"),
        (b"1234567890123456", "00", b"9999", "8888"),
        ("1234567890123456", b"00", "9999", b"8888"),
        ("1234567890123456", "00", "9999", "8888"),
    ],
)
def test_visa_cvn18(
    pan: Union[bytes, str],
    psn: Union[bytes, str],
    pin: Union[bytes, str],
    current_pin: Union[bytes, str],
) -> None:
    """Visa CVN 18"""
    cvn18 = cvn.VisaCVN18(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Check ICC Master Keys
    assert key_check_digits(cvn18.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn18.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn18.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    arqc = cvn18.generate_ac(
        tag_9f02=bytes.fromhex("000000004000"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000048000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("191105"),
        tag_9c=bytes.fromhex("01"),
        tag_9f37=bytes.fromhex("52BF4585"),
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        tag_9f10=bytes.fromhex("06011203A0B800"),
    )
    assert arqc.hex().upper() == "7A788EA6B8A3E733"
    assert key_check_digits(cvn18._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # ARPC and its session key
    arpc = cvn18.generate_arpc(
        tag_9f26=arqc, tag_9f36=atc, csu=bytes.fromhex("00000000")
    )
    assert arpc.hex().upper() == "9AF514C1"
    assert key_check_digits(cvn18._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # MAC application unblock command and its session key
    mac = cvn18.generate_command_mac(
        bytes.fromhex("8418000008"),
        tag_9f26=arqc,
        tag_9f36=atc,
    )
    assert mac.hex().upper() == "04EDE90BC24CC35E"
    assert (
        key_check_digits(cvn18._derive_sk_sm_visa(cvn18.icc_mk_smi, atc)).hex().upper()
        == "DB10"
    )

    # PIN change without current PIN and its session keys
    pin_command = cvn18.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert (
        pin_command.hex().upper()
        == "8424000218BB221AF527377A2811D2B6EBC396A9391A22B480EF312F40"
    )
    assert (
        key_check_digits(cvn18._derive_sk_sm_visa(cvn18.icc_mk_smc, atc)).hex().upper()
        == "6D04"
    )
    assert (
        key_check_digits(cvn18._derive_sk_sm_visa(cvn18.icc_mk_smi, atc)).hex().upper()
        == "DB10"
    )

    # PIN change with current PIN and its session keys
    pin_command = cvn18.generate_pin_change_command(
        pin, tag_9f26=arqc, tag_9f36=atc, current_pin=current_pin
    )
    assert (
        pin_command.hex().upper()
        == "84240001182EE025E554BCD87E11D2B6EBC396A9390FB94A4E4EC10F41"
    )
    assert (
        key_check_digits(cvn18._derive_sk_sm_visa(cvn18.icc_mk_smc, atc)).hex().upper()
        == "6D04"
    )
    assert (
        key_check_digits(cvn18._derive_sk_sm_visa(cvn18.icc_mk_smi, atc)).hex().upper()
        == "DB10"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin", "current_pin"],
    [
        (b"1234567890123456", b"00", b"9999", b"8888"),
        (b"1234567890123456", "00", b"9999", "8888"),
        ("1234567890123456", b"00", "9999", b"8888"),
        ("1234567890123456", "00", "9999", "8888"),
    ],
)
def test_visa_cvn22(
    pan: Union[bytes, str],
    psn: Union[bytes, str],
    pin: Union[bytes, str],
    current_pin: Union[bytes, str],
) -> None:
    """Visa CVN 22"""
    cvn22 = cvn.VisaCVN22(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Check ICC Master Keys
    assert key_check_digits(cvn22.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn22.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn22.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    arqc = cvn22.generate_ac(
        tag_9f02=bytes.fromhex("000000004000"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000048000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("191105"),
        tag_9c=bytes.fromhex("01"),
        tag_9f37=bytes.fromhex("52BF4585"),
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        tag_9f10=bytes.fromhex("06011203A0B800"),
    )
    assert arqc.hex().upper() == "7A788EA6B8A3E733"
    assert key_check_digits(cvn22._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # ARPC and its session key
    arpc = cvn22.generate_arpc(
        tag_9f26=arqc, tag_9f36=atc, csu=bytes.fromhex("00000000")
    )
    assert arpc.hex().upper() == "9AF514C1"
    assert key_check_digits(cvn22._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # MAC application unblock command and its session key
    mac = cvn22.generate_command_mac(
        bytes.fromhex("8418000008"),
        tag_9f26=arqc,
        tag_9f36=atc,
    )
    assert mac.hex().upper() == "A4805748F846D851"
    assert (
        key_check_digits(cvn22._derive_sk_sm_common(cvn22.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "7A15"
    )

    # PIN change without current PIN and its session keys
    pin_command = cvn22.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert (
        pin_command.hex().upper()
        == "84240002181B2943BED3D298F9C75848D17317CD1D25B70F7E26613E5F"
    )
    assert (
        key_check_digits(cvn22._derive_sk_sm_common(cvn22.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "B9FB"
    )
    assert (
        key_check_digits(cvn22._derive_sk_sm_common(cvn22.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "7A15"
    )

    # PIN change with current PIN and its session keys
    pin_command = cvn22.generate_pin_change_command(
        pin, tag_9f26=arqc, tag_9f36=atc, current_pin=current_pin
    )
    assert (
        pin_command.hex().upper()
        == "8424000118AD6CF0C44DD96B7323A1B51B694E99295A640C9E585FF059"
    )
    assert (
        key_check_digits(cvn22._derive_sk_sm_common(cvn22.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "B9FB"
    )
    assert (
        key_check_digits(cvn22._derive_sk_sm_common(cvn22.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "7A15"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin"],
    [
        (b"1234567890123456", b"00", b"9999"),
        (b"1234567890123456", "00", b"9999"),
        ("1234567890123456", b"00", "9999"),
        ("1234567890123456", "00", "9999"),
    ],
)
def test_interac_cvn133(
    pan: Union[bytes, str], psn: Union[bytes, str], pin: Union[bytes, str]
) -> None:
    """Interac CVN 133"""
    cvn133 = cvn.InteracCVN133(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Derive ICC Master Keys
    assert key_check_digits(cvn133.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn133.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn133.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    un = bytes.fromhex("ABCDEF12")
    arqc = cvn133.generate_ac(
        tag_9f02=bytes.fromhex("000000009999"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000000000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("209906"),
        tag_9c=bytes.fromhex("00"),
        tag_9f37=un,
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        tag_9f10=bytes.fromhex("1501850440B100000000B00000000000000000000000"),
    )
    assert arqc.hex().upper() == "18932FECA2D84326"
    assert (
        key_check_digits(cvn133._derive_sk_ac_mastercard(atc, un)).hex().upper()
        == "2FD5"
    )

    # ARPC and its session key
    arpc = cvn133.generate_arpc(
        tag_9f26=arqc, tag_9f37=un, tag_9f36=atc, arpc_rc=bytes.fromhex("0000")
    )
    assert arpc.hex().upper() == "4DF171A49396C363"
    assert (
        key_check_digits(cvn133._derive_sk_ac_mastercard(atc, un)).hex().upper()
        == "2FD5"
    )

    # MAC PIN unblock command and its session key
    mac = cvn133.generate_command_mac(bytes.fromhex("8424000008"), tag_9f26=arqc)
    assert mac.hex().upper() == "B574ABB0A485A330"
    assert (
        key_check_digits(cvn133._derive_sk_sm_common(cvn133.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "08DF"
    )

    # PIN change without current PIN and its session keys
    pin_command = cvn133.generate_pin_change_command(pin, tag_9f26=arqc)
    assert pin_command.hex().upper() == "8424000210382F36D5920D23915CC7573F770298C9"
    assert (
        key_check_digits(cvn133._derive_sk_sm_common(cvn133.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "748D"
    )
    assert (
        key_check_digits(cvn133._derive_sk_sm_common(cvn133.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "08DF"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin"],
    [
        (b"1234567890123456", b"00", b"9999"),
        (b"1234567890123456", "00", b"9999"),
        ("1234567890123456", b"00", "9999"),
        ("1234567890123456", "00", "9999"),
    ],
)
def test_mastercard_cvn16(
    pan: Union[bytes, str], psn: Union[bytes, str], pin: Union[bytes, str]
) -> None:
    """MasterCard CVN 16"""
    cvn16 = cvn.MasterCardCVN16(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Derive ICC Master Keys
    assert key_check_digits(cvn16.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn16.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn16.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    un = bytes.fromhex("ABCDEF12")
    arqc = cvn16.generate_ac(
        tag_9f02=bytes.fromhex("000000009999"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000000000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("209906"),
        tag_9c=bytes.fromhex("00"),
        tag_9f37=un,
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        cvr=bytes.fromhex("A00003220000"),  # 9F10: 0110A00003220000000000000000000000FF
    )
    assert arqc.hex().upper() == "24CCF3DEE3158C70"
    assert (
        key_check_digits(cvn16._derive_sk_ac_mastercard(atc, un)).hex().upper()
        == "2FD5"
    )

    # ARPC and its session key (no session key, uses ICC MK AC)
    arpc = cvn16.generate_arpc(tag_9f26=arqc, arpc_rc=bytes.fromhex("0010"))
    assert arpc.hex().upper() == "73F6BBF389A6586C"
    assert key_check_digits(cvn16._derive_sk_arpc_none()).hex().upper() == "BAB0"

    # MAC PIN unblock command and its session key
    # 72169F180400000000860D8424000008ED0AB162BAB1DE61
    mac = cvn16.generate_command_mac(
        bytes.fromhex("8424000008"), tag_9f26=arqc, tag_9f36=atc
    )
    assert mac.hex().upper() == "ED0AB162BAB1DE61"
    assert (
        key_check_digits(cvn16._derive_sk_sm_common(cvn16.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "FFA0"
    )

    # PIN change without current PIN and its session keys
    # 721E9F1804000000008615842400021000859B8FE53F316DED0AB162BAB1DE61
    pin_command = cvn16.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert pin_command.hex().upper() == "842400021000859B8FE53F316DEB4B64C36BF88E39"
    assert (
        key_check_digits(cvn16._derive_sk_sm_common(cvn16.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "0BCA"
    )
    assert (
        key_check_digits(cvn16._derive_sk_sm_common(cvn16.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "FFA0"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin"],
    [
        (b"1234567890123456", b"00", b"9999"),
        (b"1234567890123456", "00", b"9999"),
        ("1234567890123456", b"00", "9999"),
        ("1234567890123456", "00", "9999"),
    ],
)
def test_mastercard_cvn17(
    pan: Union[bytes, str], psn: Union[bytes, str], pin: Union[bytes, str]
) -> None:
    """MasterCard CVN 17"""
    cvn17 = cvn.MasterCardCVN17(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Derive ICC Master Keys
    assert key_check_digits(cvn17.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn17.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn17.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    un = bytes.fromhex("ABCDEF12")
    arqc = cvn17.generate_ac(
        tag_9f02=bytes.fromhex("000000009999"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000000000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("209906"),
        tag_9c=bytes.fromhex("00"),
        tag_9f37=un,
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        cvr=bytes.fromhex("A00003220000"),  # 9F10: 0111A00003220000000000000000000322FF
        counters=bytes.fromhex("00000000000322FF"),
    )

    assert arqc.hex().upper() == "EAE8A620A75648D8"
    assert (
        key_check_digits(cvn17._derive_sk_ac_mastercard(atc, un)).hex().upper()
        == "2FD5"
    )

    # ARPC and its session key (no session key, uses ICC MK AC)
    arpc = cvn17.generate_arpc(tag_9f26=arqc, arpc_rc=bytes.fromhex("0010"))
    assert arpc.hex().upper() == "29CF11FBDB9CA650"
    assert key_check_digits(cvn17._derive_sk_arpc_none()).hex().upper() == "BAB0"

    # MAC PIN unblock command and its session key
    # 72169F180400000000860D84240000084D1858456CD0F22A
    mac = cvn17.generate_command_mac(
        bytes.fromhex("8424000008"), tag_9f26=arqc, tag_9f36=atc
    )
    assert mac.hex().upper() == "4D1858456CD0F22A"
    assert (
        key_check_digits(cvn17._derive_sk_sm_common(cvn17.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "6CAF"
    )

    # PIN change without current PIN and its session keys
    # 721E9F180400000000861584240002108661C474E4940A4378390C64CB96756E
    pin_command = cvn17.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert pin_command.hex().upper() == "84240002108661C474E4940A4378390C64CB96756E"
    assert (
        key_check_digits(cvn17._derive_sk_sm_common(cvn17.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "3D99"
    )
    assert (
        key_check_digits(cvn17._derive_sk_sm_common(cvn17.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "6CAF"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin"],
    [
        (b"1234567890123456", b"00", b"9999"),
        (b"1234567890123456", "00", b"9999"),
        ("1234567890123456", b"00", "9999"),
        ("1234567890123456", "00", "9999"),
    ],
)
def test_mastercard_cvn20(
    pan: Union[bytes, str], psn: Union[bytes, str], pin: Union[bytes, str]
) -> None:
    """MasterCard CVN 20"""
    cvn20 = cvn.MasterCardCVN20(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Derive ICC Master Keys
    assert key_check_digits(cvn20.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn20.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn20.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    un = bytes.fromhex("ABCDEF12")
    arqc = cvn20.generate_ac(
        tag_9f02=bytes.fromhex("000000009999"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000000000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("209906"),
        tag_9c=bytes.fromhex("00"),
        tag_9f37=un,
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        cvr=bytes.fromhex("A00003220000"),  # 9F10: 0110A00003220000000000000000000000FF
    )
    assert arqc.hex().upper() == "CD29615D6452E70E"
    assert key_check_digits(cvn20._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # ARPC and its session key
    arpc = cvn20.generate_arpc(
        tag_9f26=arqc, tag_9f36=atc, arpc_rc=bytes.fromhex("0010")
    )
    assert arpc.hex().upper() == "DFD0956606B68D64"
    assert key_check_digits(cvn20._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # MAC PIN unblock command and its session key
    # 72169F180400000000860D842400000801C0F0DFBF6AA230
    mac = cvn20.generate_command_mac(
        bytes.fromhex("8424000008"), tag_9f26=arqc, tag_9f36=atc
    )
    assert mac.hex().upper() == "01C0F0DFBF6AA230"
    assert (
        key_check_digits(cvn20._derive_sk_sm_common(cvn20.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "9240"
    )

    # PIN change without current PIN and its session keys
    # 721E9F1804000000008615842400021000859B8FE53F316DED0AB162BAB1DE61
    pin_command = cvn20.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert pin_command.hex().upper() == "84240002107647E71F85AA6A35AE4B4E838666773D"
    assert (
        key_check_digits(cvn20._derive_sk_sm_common(cvn20.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "B087"
    )
    assert (
        key_check_digits(cvn20._derive_sk_sm_common(cvn20.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "9240"
    )


@pytest.mark.parametrize(
    ["pan", "psn", "pin"],
    [
        (b"1234567890123456", b"00", b"9999"),
        (b"1234567890123456", "00", b"9999"),
        ("1234567890123456", b"00", "9999"),
        ("1234567890123456", "00", "9999"),
    ],
)
def test_mastercard_cvn21(
    pan: Union[bytes, str], psn: Union[bytes, str], pin: Union[bytes, str]
) -> None:
    """MasterCard CVN 21"""
    cvn21 = cvn.MasterCardCVN21(
        iss_mk_ac=bytes.fromhex("0123456789ABCDEFFEDCBA9876543210"),
        iss_mk_smi=bytes.fromhex("FEDCBA98765432100123456789ABCDEF"),
        iss_mk_smc=bytes.fromhex("89ABCDEF0123456776543210FEDCBA98"),
        pan=pan,
        psn=psn,
    )

    # Derive ICC Master Keys
    assert key_check_digits(cvn21.icc_mk_ac).hex().upper() == "BAB0"
    assert key_check_digits(cvn21.icc_mk_smi).hex().upper() == "F010"
    assert key_check_digits(cvn21.icc_mk_smc).hex().upper() == "B154"

    # ARQC and its session key
    atc = bytes.fromhex("001C")
    un = bytes.fromhex("ABCDEF12")
    arqc = cvn21.generate_ac(
        tag_9f02=bytes.fromhex("000000009999"),
        tag_9f03=bytes.fromhex("000000000000"),
        tag_9f1a=bytes.fromhex("0124"),
        tag_95=bytes.fromhex("8000000000"),
        tag_5f2a=bytes.fromhex("0124"),
        tag_9a=bytes.fromhex("209906"),
        tag_9c=bytes.fromhex("00"),
        tag_9f37=un,
        tag_82=bytes.fromhex("1800"),
        tag_9f36=atc,
        cvr=bytes.fromhex("A00003220000"),  # 9F10: 0115A00003220000000000000000000322FF
        counters=bytes.fromhex("00000000000322FF"),
    )

    assert arqc.hex().upper() == "C30E54BE496CF5E5"
    assert key_check_digits(cvn21._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # ARPC and its session key
    arpc = cvn21.generate_arpc(
        tag_9f26=arqc, tag_9f36=atc, arpc_rc=bytes.fromhex("0010")
    )
    assert arpc.hex().upper() == "4F51F9CC63CE8B3C"
    assert key_check_digits(cvn21._derive_sk_ac_common(atc)).hex().upper() == "22C8"

    # MAC PIN unblock command and its session key
    # 72169F180400000000860D84240000085E89E990C51CEE94
    mac = cvn21.generate_command_mac(
        bytes.fromhex("8424000008"), tag_9f26=arqc, tag_9f36=atc
    )
    assert mac.hex().upper() == "5E89E990C51CEE94"
    assert (
        key_check_digits(cvn21._derive_sk_sm_common(cvn21.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "F4C4"
    )

    # PIN change without current PIN and its session keys
    # 721E9F18040000000086158424000210D9F5896BE8E283C682264E1D32DF1D51
    pin_command = cvn21.generate_pin_change_command(pin, tag_9f26=arqc, tag_9f36=atc)
    assert pin_command.hex().upper() == "8424000210D9F5896BE8E283C682264E1D32DF1D51"
    assert (
        key_check_digits(cvn21._derive_sk_sm_common(cvn21.icc_mk_smc, arqc))
        .hex()
        .upper()
        == "BB95"
    )
    assert (
        key_check_digits(cvn21._derive_sk_sm_common(cvn21.icc_mk_smi, arqc))
        .hex()
        .upper()
        == "F4C4"
    )
